from kalib import (
    classes,
    datastructures,
    descriptors,
    importer,
    internals,
    loggers,
    misc,
    monkey,
    signals,
    text,
    versions,
)
from kalib.classes import (
    Nothing,
)
from kalib.dataclass import (
    autoclass,
    dataclass,
    to_namedtuple,
)
from kalib.datastructures import (
    dumps,
    json,
    loads,
    pack,
    serializer,
    unpack,
)
from kalib.descriptors import (
    Property,
    cache,
    pin,
)
from kalib.exceptions import (
    exception,
)
from kalib.functions import (
    to_ascii,
    to_bytes,
)
from kalib.hypertext import (
    HTTP,
)
from kalib.importer import (
    add_path,
    optional,
    required,
    sort,
)
from kalib.internals import (
    Who,
    class_of,
    is_class,
    issubstance,
    stacktrace,
    unique,
)
from kalib.loggers import (
    Logging,
    logger,
)
from kalib.misc import (
    Now,
    Timer,
    lazy_proxy_to,
    stamp,
    tty,
)
from kalib.monkey import (
    Monkey,
)
from kalib.signals import (
    quit_at,
)
from kalib.text import (
    Str,
)
from kalib.versions import (
    Git,
)

__all__ = (
    'HTTP',
    'Git',
    'Logging',
    'Monkey',
    'Nothing',
    'Now',
    'Property',
    'Str',
    'Time',
    'Who',
    'add_path',
    'autoclass',
    'cache',
    'class_of',
    'classes',
    'dataclass',
    'datastructures',
    'descriptors',
    'dumps',
    'exception',
    'importer',
    'internals',
    'is_class',
    'issubstance',
    'json',
    'lazy_proxy_to',
    'loads',
    'logger',
    'loggers',
    'misc',
    'monkey',
    'optional',
    'pack',
    'pin',
    'quit_at',
    'required',
    'serializer',
    'signals',
    'sort',
    'stacktrace',
    'stamp',
    'text',
    'to_ascii',
    'to_bytes',
    'to_namedtuple',
    'tty',
    'unique',
    'unpack',
    'versions',
)

Time = Timer()



from collections import namedtuple
from datetime import date, datetime
from traceback import format_exception

from kalib.datastructures import json
from kalib.internals import Who
from kalib.text import Str


class ExceptionInformation(namedtuple('Exception', 'arguments message reason trace')):
    __slots__ = ()

    @property
    def as_dict(self):
        return self._asdict()


def exception(e):
    def trim(x):
        return tuple(i.rstrip() for i in x)

    def select_primitives_only(e):
        if (args := getattr(e, 'args', None)):
            result = []
            for item in args:
                if not item:
                    continue

                # TODO: replace with datastructures.cast

                if isinstance(item, list | set):
                    item = tuple(item)  # noqa: PLW2901

                elif isinstance(item, date | datetime):
                    item = item.isoformat()  # noqa: PLW2901

                elif isinstance(item, bytes | float | int | str | tuple):
                    ...

                else:
                    try:
                        item = json.dumps(item)  # noqa: PLW2901
                    except Exception:  # noqa: S112, BLE001
                        continue

                result.append(item)

            if result:
                return json.dumps(result)

    arguments = select_primitives_only(e)
    message = Str(e).strip() or None

    return ExceptionInformation(
        arguments, message,
        f'{Who(e)}({arguments}) {message or ""}',
        '\n'.join(trim(format_exception(e))))



import sys
import time
from pathlib import Path
from signal import signal as bind

from kalib.descriptors import cache
from kalib.loggers import Logging

Logger = Logging.Default
NeedRestart = False


@cache
def get_selfpath():
    return Path(sys.argv[0]).resolve()


def get_mtime():
    return get_selfpath().stat().st_mtime


@cache
def quit_at(*, func=sys.exit, signal=None, errno=137, **kw):
    """Quit the program when the runned file is updated or signal is received."""

    def handler(signum, frame):  # noqa: ARG001
        global NeedRestart  # noqa: PLW0603
        NeedRestart = True
        Logger.warning(f'{signal=} received, mark restart as needed')

    if signal:
        bind(signal, handler)

    initial_stamp = get_mtime()
    def on_change(*, sleep=0.0):

        if NeedRestart and signal:
            Logger.warning(f'{signal=} received, quit..')
            func(errno)
            return False

        elif initial_stamp != (ctime := get_mtime()):
            file = get_selfpath()
            Logger.warning(
                f'{file=} updated {time.time() - ctime:.2f} seconds ago, quit..')
            func(errno)
            return False

        if sleep := (sleep or kw.get('sleep', 0.0)):
            time.sleep(sleep)
        return True
    return on_change



import sys
from contextlib import suppress
from functools import cache
from importlib import import_module
from inspect import ismodule
from os import sep
from pathlib import Path

from kalib.functions import to_ascii
from kalib.internals import Who, iter_stack, unique
from kalib.loggers import Logging

__all__ = 'add_path', 'optional', 'required'

logger = Logging.get(__name__)


IGNORED_OBJECT_FIELDS = {
    '__builtins__', '__cached__', '__doc__', '__file__', '__loader__',
    '__name__', '__package__', '__path__', '__spec__'}
PACKAGES_MAP = {'magic': 'python-magic', 'git': 'gitpython'}


@cache
def get_module(path):

    chunks = path.split('.')
    count = len(chunks) + 1

    if count == 2:  # noqa: PLR2004
        with suppress(ModuleNotFoundError):
            return import_module(path), ()

    for i in range(1, count):
        chunk = '.'.join(chunks[:count - i])
        with suppress(ModuleNotFoundError):
            return import_module(chunk), tuple(chunks[count - i:])

    msg = f"ImportError: {path} ({chunk!a} isn't exists)"
    raise ImportError(msg)


def get_child(path, parent, child):
    if ismodule(parent):
        __import__(parent.__name__, globals(), locals(), [str(child)])

    if not hasattr(parent, child):
        from kalib.misc import sourcefile

        if not ismodule(parent):
            raise ImportError(
                f"{path} (object {Who(parent)!a} hasn't attribute "
                f"{child!a}{sourcefile(parent, 'in %a')})")

        if not set(dir(parent)) - IGNORED_OBJECT_FIELDS:
            chunk = f'{Who(parent)}.{child}'
            raise ImportError(
                f'{path} (from partially initialized module '
                f'{chunk!a}, most likely due to a circular import'
                f'{sourcefile(parent, "from %a")}) or just not found')

        raise ImportError(
            f"{path} (module {Who(parent)!a} hasn't member {child!a}"
            f"{sourcefile(parent, 'in %a')})")

    return getattr(parent, child)


def import_object(path, something=None):
    if path is something is None:
        raise TypeError('all arguments is None')

    if isinstance(path, str | bytes):
        path = to_ascii(path)

    if not isinstance(path, str):
        if something is None:
            msg = (
                f"{Who.Is(path)} isn't str, but "
                f'second argument (import path) is None')
            raise TypeError(msg)
        path, something = something, path

    logger.debug(f'lookup: {path}')

    if something:
        locator = f'{Who(something)}.{path}'
        sequence = path.split('.')

    else:
        locator = f'{path}'
        something, sequence = get_module(path)

        if something is None:
            raise ImportError(f"{path} (isn't exists?)")

    if not sequence:
        logger.debug(f'import path: {Who(something)}')

    else:
        logger.debug(
            f'split path: {Who(something)} '
            f'(module) -> {".".join(sequence)} (path)')

    for name in sequence:
        something = get_child(locator, something, name)

    logger.debug('load ok: %s', path)
    return something


@cache
def cached_import(*args, **kw):
    return import_object(*args, **kw)


def required(path, *args, **kw):
    """For dynamic import any from any, usage:

    required('kalib.importer.required')  # import kalib.importer and return function

    """

    throw = kw.pop('throw', True)
    quiet = kw.pop('quiet', False)
    default = kw.pop('default', None)

    def wrap_uncacheable(*args, **kw):
        try:
            return cached_import(*args, **kw)

        except TypeError:
            return import_object(*args, **kw)

    try:
        return wrap_uncacheable(path, *args, **kw)

    except ImportError as e:

        if not quiet or throw:
            msg = f"couldn't import required({path=}, *{args=}, **{kw=})"

            base = path.split('.', 1)[0]
            if base not in sys.modules:
                package = (PACKAGES_MAP.get(base) or base).replace('_', '-')
                msg = f'{msg}; (need extra {package=}?)'

            if not quiet:
                Logging.Default.warning(msg, trace=True)

            if throw:
                raise ImportError(msg) from e

    return default


def optional(path, *args, **kw):
    kw.setdefault('quiet', True)
    kw.setdefault('throw', False)
    return required(path, *args, **kw)


sort = optional('natsort.natsorted', quiet=True, default=sorted)


def get_path(path, root=None):  # noqa: PLR0912

    if root is None:

        base = Path(__file__).stem
        for file in iter_stack(1, offset=1):
            if Path(file).stem != base:
                break
        root = Path(file).parent

    if isinstance(root, Path | str):
        root = Path(root)
    else:
        raise TypeError(
            f'root={root!r} can be str | {Who(Path)} | None, not {Who.Is(root)}')

    spath = str(path).strip('/')

    if set(spath) == {'.'}:
        dots = len(spath) - 2
        if dots == -1:
            return path

        path = root.resolve()
        for _ in range(dots + 1):
            path = path.parent

        return path.resolve()

    elif spath.startswith('../'):
        return (root / path).resolve()

    elif sep in path and ('../' not in spath and '/..' not in spath):
        try:
            idx = str(root).index(str(path))
        except ValueError as e:
            raise ValueError(f'{path=} not found in {root=}') from e
        return Path(str(root)[:idx])

    subdir = str(root)
    while subdir != sep:

        future = Path(subdir)
        subdir = str(future.parent)

        if path == future.name:
            return future

    raise ValueError(f'{path=} not found in {root=}')


def add_path(path, **kw):
    """Add resolved path to sys.path and return Path object, usage:

    add_path('..')      # ../
    add_path('...')     # ../../
    add_path('....')    # ../../../

    app_path(__file__)  # add current file directory to sys.path
    app_path('/etc/hosts')  # add /etc to sys.path, hosts is file

    """

    path = Path(path)
    request = path

    if path.is_file():
        path = path.resolve().parent

    elif not (str(path).startswith(sep) or path == path.resolve()):
        root = get_path(path, **kw)
        if not root:
            raise ValueError(f'{path=} not found, {logger.Args(**kw)}')
        path = root if str(path).startswith('.') else (root / path).resolve()

    if path not in sys.path:
        sys.path.append(str(path.resolve()))
        sys.path = list(unique(sys.path))
        exists = path.is_dir()
        logger.info(f'path {request} resolved to {path}, {exists=}')
    return path



from kalib.internals import Who

__all__ = 'Nothing', 'Singleton'


class Singleton(type):

    def __init__(cls, name, parents, attrbutes):
        super().__init__(name, parents, attrbutes)

    def __call__(cls, *args, **kw):
        if not hasattr(cls, 'instance'):
            cls.instance = super().__call__(*args, **kw)
        return cls.instance


class Missing(metaclass=Singleton):
    def __bool__(self):
        return False

    def __repr__(self):
        return f'<{Who.Name(self)}>'


# This is a singleton, so it's safe to use it as a sentinel

Nothing = Missing()



"""Using for halt the program with error code 127, usage, just:

from kalib import halt
"""

import sys

from kalib.loggers import Logging

Logging.get(__name__).exception('halt', trace=True, shift=-2, stack=2)
sys.exit(127)
